/* Copyright (C) 2011 TigerVNC Team
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/*
 * This file contains resource IDs shared between various Windows binaries.
 * If you are adding new resource ID ensure it doesn't conflict with per-binary
 * resource IDs.
 */

#ifndef __RFB_WIN32_RESOURCE_H__
#define __RFB_WIN32_RESOURCE_H__

#define IDD_SECURITY		117

#define IDC_VENCRYPT		1200
#define IDC_ENC_NONE		1201
#define IDC_ENC_TLS		1202
#define IDC_ENC_X509		1203
#define IDC_AUTH_NONE		1206
#define IDC_AUTH_VNC		1207
#define IDC_AUTH_PLAIN		1208

#endif
